/*
 * linux/include/asm-arm/arch-ebsa110/mm-init.h
 *
 * Copyright (C) 1997,1998 Russell King
 *
 * Description of the initial memory map for EBSA-110
 */

static init_mem_map_t init_mem_map[] = {
	INIT_MEM_MAP_SENTINEL
};
