/*
 * linux/include/asm-arm/arch-arc/mmu.h
 *
 * Copyright (c) 1996 Russell King.
 *
 * Changelog:
 *  22-11-1996	RMK	Created
 */
#ifndef __ASM_ARCH_MMU_H
#define __ASM_ARCH_MMU_H

#define __virt_to_phys(vpage) vpage
#define __phys_to_virt(ppage) ppage

#endif
