/*
 * linux/include/asm-arm/timex.h
 *
 * Architecture Specific TIME specifications
 *
 * Copyright (C) 1997,1998 Russell King
 */
#ifndef _ASMARM_TIMEX_H
#define _ASMARM_TIMEX_H

#include <asm/arch/timex.h>

#endif
