/*
 * *** THIS FILE IS AUTOMATICALLY GENERATED.  DO NOT EDIT! ***
 */
unsigned long addr_limit = 56;
#define __HAS_addr_limit
unsigned long tss_memmap = 640;
#define __HAS_tss_memmap
unsigned long mm = 1676;
#define __HAS_mm
unsigned long pgd = 8;
#define __HAS_pgd
unsigned long tss_save = 636;
#define __HAS_tss_save
unsigned long tss_fpesave = 492;
#define __HAS_tss_fpesave
unsigned long tss_memcmap = 644;
#define __HAS_tss_memcmap
