/*
 * linux/include/asm-arm/arch-a5k/timex.h
 *
 * A5000 architecture timex specifications
 *
 * Copyright (C) 1997, 1998 Russell King
 */

/*
 * On the RiscPC, the clock ticks at 2MHz.
 */
#define CLOCK_TICK_RATE		2000000

